# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Terminal Plugin."""

# Standard imports
import functools
import os
import os.path as osp
import requests
import sys

# Third party imports
import qstylizer
from qtpy.QtCore import QProcess, QTimer, Signal, Slot
from qtpy.QtWidgets import QMessageBox, QVBoxLayout
from spyder.api.config.decorators import on_conf_change
from spyder.api.widgets.main_widget import PluginMainWidget
from spyder.config.base import get_translation, get_debug_level
from spyder.utils.misc import select_port
from spyder.utils.programs import find_program
from spyder.widgets.tabs import Tabs
from spyder.plugins.remoteclient.api.protocol import ClientType

# Local imports
from spyder_terminal.api import (
    TerminalMainWidgetActions,
    TermViewMenus,
    TerminalMainWidgetMenus,
)
from spyder_terminal.widgets.findwidget import FindTerminal
from spyder_terminal.widgets.terminalgui import TerminalWidget


# Constants
LOCATION = osp.realpath(osp.join(os.getcwd(), osp.dirname(__file__)))
WINDOWS = os.name == "nt"

# For translations
_ = get_translation("spyder_terminal")


class TerminalMainWidgetToolbarSections:
    New = "new"
    Reload = "reload"


class TerminalMainWidgetCornerToolbar:
    NewTerm = "new_terminal"


class TerminalMainWidgetMenuSections:
    New = "new"
    TabActions = "tab_actions"


class TerminalMainWidget(PluginMainWidget):
    """
    Terminal plugin main widget.
    """

    MAX_SERVER_CONTACT_RETRIES = 40
    URL_ISSUES = " https://github.com/spyder-ide/spyder-terminal/issues"

    # --- Signals
    # ------------------------------------------------------------------------
    sig_server_is_ready = Signal()
    """
    This signal is emitted when the server is ready to connect.
    """

    def __init__(self, name, plugin, parent):
        """Widget constructor."""
        self.terms = []
        super().__init__(name, plugin, parent)

        # Attributes
        self.tab_widget = None
        self.menu_actions = None
        self.server_retries = 0
        self.server_ready = False
        self.font = None
        self.port = select_port(default_port=8071)
        self.stdout_file = None
        self.stderr_file = None
        if get_debug_level() > 0:
            self.stdout_file = osp.join(os.getcwd(), "spyder_terminal_out.log")
            self.stderr_file = osp.join(os.getcwd(), "spyder_terminal_err.log")
        self.project_path = None
        self.current_file_path = None
        self.current_cwd = os.getcwd()
        self.closing = False

        # Widgets
        self.main = parent
        self.find_widget = FindTerminal(self)
        self.find_widget.hide()

        layout = QVBoxLayout()

        # Tab Widget
        self.tabwidget = Tabs(self, rename_tabs=True)
        self.tabwidget.currentChanged.connect(self.refresh_plugin)
        self.tabwidget.move_data.connect(self.move_tab)
        self.tabwidget.set_close_function(self.close_term)

        if (
            hasattr(self.tabwidget, "setDocumentMode")
            and not sys.platform == "darwin"
        ):
            # Don't set document mode to true on OSX because it generates
            # a crash when the console is detached from the main window
            # Fixes Issue 561
            self.tabwidget.setDocumentMode(True)
        layout.addWidget(self.tabwidget)
        layout.addWidget(self.find_widget)
        self.setLayout(layout)

        css = qstylizer.style.StyleSheet()
        css.QTabWidget.pane.setValues(border=0)
        self.setStyleSheet(css.toString())

        self.__wait_server_to_start()

    # ---- PluginMainWidget API
    # ------------------------------------------------------------------------
    def get_focus_widget(self):
        """
        Set focus on current selected terminal.

        Return the widget to give focus to when
        this plugin's dockwidget is raised on top-level.
        """
        term = self.tabwidget.currentWidget()
        if term is not None:
            return term.view

    def get_title(self):
        """Define the title of the widget."""
        return _("Terminal")

    def setup(self):
        """Perform the setup of plugin's main menu and signals."""
        self.cmd = find_program(self.get_conf("shell"))
        server_args = [
            sys.executable,
            "-m",
            "spyder_terminal.server",
            "--port",
            str(self.port),
            "--shell",
            self.cmd,
        ]
        self.server = QProcess(self)
        env = self.server.processEnvironment()
        for var in os.environ:
            env.insert(var, os.environ[var])
        self.server.setProcessEnvironment(env)
        self.server.errorOccurred.connect(self.handle_process_errors)
        self.server.setProcessChannelMode(QProcess.SeparateChannels)
        if self.stdout_file and self.stderr_file:
            self.server.setStandardOutputFile(self.stdout_file)
            self.server.setStandardErrorFile(self.stderr_file)
        self.server.start(server_args[0], server_args[1:])
        self.color_scheme = self.get_conf("appearance", "ui_theme")
        self.theme = self.get_conf("appearance", "selected")

        # Menu
        menu = self.get_options_menu()

        # Actions
        new_terminal_toolbar_action = self.create_toolbutton(
            TerminalMainWidgetToolbarSections.New,
            text=_("Open a new terminal"),
            icon=self.create_icon("expand_selection"),
            triggered=lambda: self.create_new_term(),
        )
        reload_terminal_toolbar_action = self.create_toolbutton(
            TerminalMainWidgetToolbarSections.Reload,
            text=_("Reload terminal"),
            icon=self.create_icon("restart"),
            triggered=lambda: self.reload_current_term(),
        )

        for toolbar_action in [
            new_terminal_toolbar_action,
            reload_terminal_toolbar_action,
        ]:
            self.add_corner_widget(toolbar_action)

        new_terminal_cwd = self.create_action(
            TerminalMainWidgetActions.NewTerminalForCWD,
            text=_("New terminal in current working directory"),
            tip=_("Sets the pwd at the current working directory"),
            triggered=lambda: self.create_new_term(),
            shortcut_context="terminal",
            register_shortcut=True,
        )

        self.new_terminal_project = self.create_action(
            TerminalMainWidgetActions.NewTerminalForProject,
            text=_("New terminal in current project"),
            tip=_("Sets the pwd at the current project directory"),
            triggered=lambda: self.create_new_term(path=self.project_path),
        )

        new_terminal_file = self.create_action(
            TerminalMainWidgetActions.NewTerminalForFile,
            text=_("New terminal in current Editor file"),
            tip=_(
                "Sets the pwd at the directory that contains the current "
                "opened file"
            ),
            triggered=lambda: self.create_new_term(
                path=self.current_file_path
            ),
        )

        rename_tab_action = self.create_action(
            TerminalMainWidgetActions.RenameTab,
            text=_("Rename terminal"),
            triggered=lambda: self.tab_name_editor(),
        )

        # Context menu actions
        self.create_action(
            TerminalMainWidgetActions.Copy,
            text=_("Copy text"),
            icon=self.create_icon("editcopy"),
            shortcut_context="terminal",
            triggered=lambda: self.copy(),
            register_shortcut=True,
        )

        self.create_action(
            TerminalMainWidgetActions.Paste,
            text=_("Paste text"),
            icon=self.create_icon("editpaste"),
            shortcut_context="terminal",
            triggered=lambda: self.paste(),
            register_shortcut=True,
        )

        self.create_action(
            TerminalMainWidgetActions.Clear,
            text=_("Clear terminal"),
            shortcut_context="terminal",
            triggered=lambda: self.clear(),
            register_shortcut=True,
        )

        self.create_action(
            TerminalMainWidgetActions.ZoomIn,
            text=_("Zoom in"),
            shortcut_context="terminal",
            triggered=lambda: self.increase_font(),
            register_shortcut=True,
        )

        self.create_action(
            TerminalMainWidgetActions.ZoomOut,
            text=_("Zoom out"),
            shortcut_context="terminal",
            triggered=lambda: self.decrease_font(),
            register_shortcut=True,
        )

        # Create context menu
        self.create_menu(TermViewMenus.Context)

        # Create remote terminals menu
        self.remote_menu = self.create_menu(
            TerminalMainWidgetMenus.Remote,
            title=_("New terminal in remote server"),
        )

        # Add actions to options menu
        for item in [
            new_terminal_cwd,
            self.new_terminal_project,
            new_terminal_file,
            self.remote_menu,
        ]:
            self.add_item_to_menu(
                item, menu=menu, section=TerminalMainWidgetMenuSections.New
            )

        self.add_item_to_menu(
            rename_tab_action,
            menu=menu,
            section=TerminalMainWidgetMenuSections.TabActions,
        )

    def setup_remote_terminals_menu(self):
        self.remote_menu.clear_actions()

        for config_id in self._plugin._remote_client.get_config_ids():
            client_type = self.get_conf(
                option=f"{config_id}/client_type", section="remoteclient"
            )
            name = self._plugin._remote_client.get_server_name(config_id)

            if client_type != ClientType.SSH:
                continue

            action = self.create_action(
                name=config_id,
                text=f"New terminal in {name} server",
                triggered=functools.partial(
                    self.create_new_term,
                    config_id=config_id,
                ),
                overwrite=True,
            )
            self.add_item_to_menu(action, menu=self.remote_menu)

    def update_actions(self):
        """Setup and update the actions in the options menu."""
        if self.project_path is None:
            self.new_terminal_project.setEnabled(False)

    # ------ Private API ------------------------------------------
    def copy(self):
        if self.get_focus_widget():
            self.get_focus_widget().copy()

    def paste(self):
        if self.get_focus_widget():
            self.get_focus_widget().paste()

    def clear(self):
        if self.get_focus_widget():
            self.get_focus_widget().clear()

    def increase_font(self):
        if self.get_focus_widget():
            self.get_focus_widget().increase_font()

    def decrease_font(self):
        if self.get_focus_widget():
            self.get_focus_widget().decrease_font()

    def __wait_server_to_start(self):
        try:
            code = requests.get(
                "http://127.0.0.1:{0}".format(self.port)
            ).status_code
        except:
            code = 500

        if self.server_retries == self.MAX_SERVER_CONTACT_RETRIES:
            QMessageBox.critical(
                self,
                _("Spyder Terminal Error"),
                _(
                    "Terminal server could not be located at "
                    '<a href="http://127.0.0.1:{0}">'
                    "http://127.0.0.1:{0}</a>, "
                    "please restart Spyder on debug mode "
                    "and open an issue with the contents of "
                    "<tt>{1}</tt> and <tt>{2}</tt> "
                    "files at {3}."
                ).format(
                    self.port,
                    self.stdout_file,
                    self.stderr_file,
                    self.URL_ISSUES,
                ),
                QMessageBox.Ok,
            )
        elif code != 200:
            self.server_retries += 1
            QTimer.singleShot(250, self.__wait_server_to_start)
        elif code == 200:
            self.sig_server_is_ready.emit()
            self.server_ready = True
            self.create_new_term(give_focus=False)

    # ------ Plugin API --------------------------------
    def update_font(self, font):
        """Update font from Preferences."""
        self.font = font
        for term in self.terms:
            term.set_font(font.family())

    def on_close(self, cancelable=False):
        """Perform actions before parent main window is closed."""
        self.closing = True
        for term in self.terms:
            try:
                term.terminal_closed.disconnect()
            except TypeError:
                pass
            term.close()
        self.server.kill()
        self.server.waitForFinished()
        return True

    def refresh_plugin(self):
        """Refresh tabwidget."""
        term = None
        if self.tabwidget.count():
            term = self.tabwidget.currentWidget()
            term.view.setFocus()
        else:
            term = None

    @on_conf_change
    def apply_plugin_settings(self, options):
        """Apply the config settings."""
        term_options = {}
        for option in options:
            if option == "color_scheme_name":
                term_options[option] = option
            else:
                term_options[option] = self.get_conf(option)
        for term in self.get_terms():
            term.apply_settings(term_options)

    # ------ Public API (for terminals) -------------------------
    def get_terms(self):
        """Return terminal list."""
        return [cl for cl in self.terms if isinstance(cl, TerminalWidget)]

    def get_current_term(self):
        """Return the currently selected terminal."""
        try:
            terminal = self.tabwidget.currentWidget()
        except AttributeError:
            terminal = None
        if terminal is not None:
            return terminal

    def reload_current_term(self):
        """Reload currently selected terminal."""
        current_term = self.get_current_term()
        if current_term:
            current_term.reload()

    def create_new_term(
        self, name=None, give_focus=True, path=None, config_id=None
    ):
        """Add a new terminal tab."""
        if path is None:
            path = self.current_cwd
        if self.project_path is not None:
            path = self.project_path
        path = path.replace("\\", "/")
        term = TerminalWidget(
            self,
            self.port,
            path=path,
            font=self.font.family(),
            theme=self.theme,
            color_scheme=self.color_scheme,
            config_id=config_id,
        )
        self.add_tab(term)
        term.terminal_closed.connect(lambda: self.close_term(term=term))

    def close_term(self, index=None, term=None):
        """Close a terminal tab."""
        if not self.tabwidget.count():
            return
        if term is not None:
            index = self.tabwidget.indexOf(term)
        if index is None and term is None:
            index = self.tabwidget.currentIndex()
        if index is not None:
            term = self.tabwidget.widget(index)
        if term:
            term.terminal_closed.disconnect()
            term.close()
        self.tabwidget.removeTab(self.tabwidget.indexOf(term))
        if term in self.terms:
            self.terms.remove(term)
        if self.tabwidget.count() == 0:
            self.create_new_term()

    def set_project_path(self, path):
        """Refresh current project path."""
        self.project_path = path
        self.new_terminal_project.setEnabled(True)

    def set_current_opened_file(self, path):
        """Get path of current opened file in editor."""
        self.current_file_path = osp.dirname(path)

    def unset_project_path(self):
        """Refresh current project path."""
        self.project_path = None
        self.new_terminal_project.setEnabled(False)

    @Slot(str)
    def set_current_cwd(self, cwd):
        """Update current working directory."""
        self.current_cwd = cwd

    def server_is_ready(self):
        """Return server status."""
        return self.server_ready

    def search_next(self, text, case=False, regex=False, word=False):
        """Search in the current terminal for the given regex."""
        term = self.get_current_term()
        if term:
            term.search_next(text, case=case, regex=regex, word=word)

    def search_previous(self, text, case=False, regex=False, word=False):
        """Search in the current terminal for the given regex."""
        term = self.get_current_term()
        if term:
            term.search_previous(text, case=case, regex=regex, word=word)

    # ------ Public API (for tabs) ---------------------------
    def add_tab(self, widget):
        """Add tab."""
        self.terms.append(widget)
        num_term = self.tabwidget.count() + 1
        index = self.tabwidget.addTab(widget, "Terminal {0}".format(num_term))
        self.tabwidget.setCurrentIndex(index)
        self.tabwidget.setTabToolTip(index, "Terminal {0}".format(num_term))
        self.activateWindow()
        widget.view.setFocus()

    def move_tab(self, index_from, index_to):
        """
        Move tab (tabs themselves have already been moved by the tabwidget).

        Allows to change order of tabs.
        """
        term = self.terms.pop(index_from)
        self.terms.insert(index_to, term)

    def tab_name_editor(self):
        """Trigger the tab name editor."""
        index = self.tabwidget.currentIndex()
        self.tabwidget.tabBar().tab_name_editor.edit_tab(index)

    def handle_process_errors(self):
        """Handle when an error ocurrs in the server."""
        if not self.closing:
            QMessageBox.warning(
                self,
                _("Spyder Terminal Server Error"),
                _(
                    "The server that creates terminals failed to start. Please "
                    "restart Spyder in a system terminal with the command <tt> "
                    "spyder --debug-info minimal</tt> and open an issue with "
                    "the contents of <tt>{0}</tt> and <tt>{1}</tt> files at {2}."
                ).format(
                    osp.join(os.getcwd(), "spyder_terminal_out.log"),
                    osp.join(os.getcwd(), "spyder_terminal_err.log"),
                    self.URL_ISSUES,
                ),
                QMessageBox.Ok,
            )


def test():
    """Plugin visual test."""
    from spyder.utils.qthelpers import qapplication
    from unittest.mock import MagicMock

    app = qapplication(test_time=8)
    plugin_mock = MagicMock()
    term = TerminalMainWidget("terminal", plugin_mock, None)
    term.resize(900, 700)
    term._setup()
    term.setup()
    term.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Spyder Project Contributors
# Copyright (c) 2015-, Jupyter Development Team.
# Copyright (c) 2008-2015, IPython Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# -----------------------------------------------------------------------------

"""
General setup rules to download external JS dependencies
via Bower

Some functions were taken from the Jupyter Notebook
setupbase definition
See: https://github.com/jupyter/notebook/blob/master/setupbase.py
"""

import logging
import os
import os.path as osp
import shlex
import shutil
import sys

from setuptools import Command
from setuptools.command.develop import develop
from setuptools.command.sdist import sdist
from subprocess import check_call

log = logging.getLogger()

if sys.platform == 'win32':
    from subprocess import list2cmdline
else:
    def list2cmdline(cmd_list):
        return ' '.join(map(shlex.quote, cmd_list))


HERE = os.path.abspath(os.path.dirname(__file__))
COMPONENTS = osp.join(HERE, 'spyder_terminal', 'server', 'static',
                      'components')


repo_root = os.path.dirname(os.path.abspath(__file__))


def run(cmd, *args, **kwargs):
    """Echo a command before running it"""
    log.info('> ' + list2cmdline(cmd))
    kwargs['shell'] = (sys.platform == 'win32')
    return check_call(cmd, *args, **kwargs)


class BuildStatic(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        if not osp.isdir(COMPONENTS):
            log.info("running [yarn install]")
            run(['yarn', 'install'], cwd=repo_root)
            log.info("installing webpack")
            run(['npm', 'i', '-D', 'webpack'], cwd=repo_root)
            run(['npm', 'i', 'webpack-cli', 'html-webpack-plugin', '@xterm/xterm',
                 '@xterm/addon-attach', '@xterm/addon-search',
                 '@xterm/addon-web-links', '@xterm/addon-fit'],
                cwd=repo_root)
            log.info('running webpack')
            run(['npm', 'run', 'webpack'], cwd=repo_root)
            log.info("copying static assests (jquery and xterm)")
            shutil.copytree('./node_modules/jquery', './spyder_terminal/server/static/components/jquery', dirs_exist_ok=True)
            shutil.copytree('./node_modules/@xterm', './spyder_terminal/server/static/components/@xterm', dirs_exist_ok=True)


class DevelopWithBuildStatic(develop):
    def install_for_development(self):
        self.run_command('build_static')
        return develop.install_for_development(self)


class SdistWithBuildStatic(sdist):
    def run(self):
        self.run_command('build_static')
        sdist.run(self)

    def make_distribution(self):
        if not osp.isdir(COMPONENTS):
            print("\nWARNING: Server components are missing!! We can't "
                  "proceed further!\n")
            sys.exit(1)
        return sdist.make_distribution(self)


class CleanComponents(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        log.info("Removing server components")
        shutil.rmtree(COMPONENTS, ignore_errors=True)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Terminal Widget."""
# Standard library imports
import json
import os
import sys

# Third-party imports
import qstylizer
from qtpy.QtCore import (Qt, QUrl, Slot, QEvent, QTimer, Signal,
                         QObject)
from qtpy.QtGui import QKeySequence
from qtpy.QtWebChannel import QWebChannel
from qtpy.QtWebEngineWidgets import (
    QWebEnginePage,
    QWebEngineView,
    QWebEngineProfile,
)
from qtpy.QtWidgets import QFrame, QVBoxLayout, QApplication
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.api.config.decorators import on_conf_change
from spyder.config.base import get_translation
from spyder.config.gui import is_dark_interface
from spyder.plugins.remoteclient.widgets import AuthenticationMethod
from spyder.utils.palette import SpyderPalette

# Local imports
from spyder_terminal.api import TerminalMainWidgetActions, TermViewMenus
from spyder_terminal.widgets.style.themes import ANSI_COLORS


PREFIX = 'spyder_terminal.default.'

# For translations
_ = get_translation('spyder_terminal')


class TermViewSections:
    CommonActions = 'common_actions'
    ZoomActions = 'zoom_actions'


class ChannelHandler(QObject):
    """QWebChannel handler for JS calls."""

    sig_ready = Signal()
    sig_closed = Signal()

    def __init__(self, parent):
        """Handler main constructor."""
        QObject.__init__(self, parent)

    @Slot()
    def ready(self):
        """Invoke signal when terminal prompt is ready."""
        self.sig_ready.emit()

    @Slot()
    def close(self):
        """Invoke signal when terminal process was closed externally."""
        self.sig_closed.emit()


class TerminalWidget(QFrame, SpyderWidgetMixin):
    """Terminal widget."""
    zdotdir = os.environ.get('ZDOTDIR') or os.environ.get('HOME')  # what about windows?
    ENV_ROUTES = {
        "bash": ["/etc/profile", "~/.bash_profile"],
        "zsh": ["/etc/zshenv", f"{zdotdir}/.zshenv", "/etc/zprofile",
                f"{zdotdir}/.zprofile", "/etc/zshrc", f"{zdotdir}/.zshrc",
                "/etc/zlogin", f"{zdotdir}/.zlogin"],
        "fish": ["~/.config/fish/config.fish"],
        "sh": ["~/.profile", "~/.shrc", "~/.shinit"],
        "ksh": ["~/.profile", "~/.kshrc"],
        "csh": ["~/.cshrc", "~/.login"],
        "pwsh": [],
        "rbash": ["~/.bashrc", "~/.bash_profile"],
        "dash": ["~/.profile"],
        "screen": [],
        "tmux": [],
        "tcsh": ["~/.tcshrc"],
        "xonsh": ["~/.xonshrc"]
    }

    terminal_closed = Signal()
    terminal_ready = Signal()

    def __init__(
        self,
        parent,
        port,
        path='~',
        font=None,
        theme=None,
        color_scheme=None,
        config_id=None,
    ):
        """Frame main constructor."""
        super().__init__(parent, class_parent=parent)
        url = 'http://127.0.0.1:{0}?path={1}'.format(port, path)
        self.handler = ChannelHandler(self)
        self.handler.sig_ready.connect(lambda: self.terminal_ready.emit())
        self.handler.sig_closed.connect(lambda: self.terminal_closed.emit())
        self.view_ready = False
        self.view = TermView(self, term_url=url, handler=self.handler)
        self.font = font
        self.initial_path = path
        self.theme = theme
        self.color_scheme = color_scheme
        self.parent = parent
        layout = QVBoxLayout()
        layout.addWidget(self.view)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setFrameStyle(QFrame.NoFrame | QFrame.Plain)
        self.setLayout(layout)

        self.body = self.view.document

        self.handler.sig_ready.connect(lambda: self.setup_term(config_id=config_id))
        self.view.sig_focus_in_event.connect(
            lambda: self._apply_stylesheet(focus=True))
        self.view.sig_focus_out_event.connect(
            lambda: self._apply_stylesheet(focus=False))
        self._apply_stylesheet()

    def setup_term(self, config_id=None):
        """Setup other terminal options after page has loaded."""
        # This forces to display the black background
        print("\0", end='')
        self.view_ready = True
        self.set_font(self.font)
        self.set_dir(self.initial_path)
        self.current_theme = self.set_theme({})
        self.set_scrollbar_style()
        options = self.get_conf_options()
        dict_options = {}
        for option in options:
            dict_options[option] = self.get_conf(option)
        self.apply_settings(dict_options)
        self.apply_zoom()
        shell_name = self.get_conf('shell')
        if os.name != 'nt':
            # Find environment variables in the home directory given the actual
            # shell
            env_route = self.ENV_ROUTES[shell_name]
            for act_file in env_route:
                if os.path.exists(os.path.expanduser(act_file)):
                    self.exec_cmd(f"source {act_file}")
                    self.exec_cmd("clear")

        if config_id:
            # Attempt to initialize a SSH connection following remote client
            # config id
            _pass = None
            auth_method = self.get_conf(
                option=f"{config_id}/auth_method", section="remoteclient"
            )
            address = self.get_conf(
                option=f"{config_id}/{auth_method}/address",
                section="remoteclient",
            )
            username = self.get_conf(
                option=f"{config_id}/{auth_method}/username",
                section="remoteclient",
            )
            port = self.get_conf(
                option=f"{config_id}/{auth_method}/port",
                section="remoteclient",
            )
            ssh_command = (
                f"ssh {username}@{address} -o StrictHostKeychecking=no"
            )

            if port:
                ssh_command = f"{ssh_command} -p {port}"
            if auth_method == AuthenticationMethod.Password:
                _pass = self.get_conf(
                    option=f"{config_id}/password",
                    section="remoteclient",
                    secure=True,
                )
            elif auth_method == AuthenticationMethod.KeyFile:
                keyfile = self.get_conf(
                    option=f"{config_id}/keyfile", section="remoteclient"
                )
                _pass = self.get_conf(
                    option=f"{config_id}/passphrase",
                    section="remoteclient",
                    secure=True,
                )
                ssh_command = f"{ssh_command} -i {keyfile}"

            self.exec_cmd(ssh_command)
            if _pass:
                self.exec_delayed_cmd(_pass, "pass")

    def get_shortcut_data(self):
        """
        Return shortcut data, a list of tuples (shortcut, text, default).

        shortcut (QShortcut or QAction instance)
        text (string): action/shortcut description
        default (string): default key sequence
        """
        return self.view.get_shortcut_data()

    def eval_javascript(self, script):
        """Evaluate Javascript instructions inside view."""
        if self.view_ready:
            return self.view.eval_javascript(script)

    def set_scrollbar_style(self):
        """Set terminal scrollbar style."""
        if is_dark_interface():
            self.eval_javascript('addClassStyleToContainer("dark-scroll")')

    def set_dir(self, path):
        """Set terminal initial current working directory."""
        self.eval_javascript('setcwd("{0}")'.format(path))

    def set_font(self, font):
        """Set terminal font via CSS."""
        self.font = font
        self.eval_javascript('fitFont("{0}")'.format(self.font))

    @on_conf_change(section='appearance')
    def set_theme(self, _values):
        """Set theme for the terminal."""
        supported_themes = ANSI_COLORS
        new_theme = {}
        theme = self.get_conf('selected', section='appearance')
        color_scheme = self.get_conf('ui_theme', section='appearance')
        if theme not in supported_themes:
            theme = 'spyder' if color_scheme == 'light' else 'spyder/dark'
        new_theme['background'] = self.get_conf(
            '{}/background'.format(theme), section='appearance')
        new_theme['foreground'] = self.get_conf(
            '{}/normal'.format(theme), section='appearance')[0]
        new_theme['cursor'] = self.get_conf(
            '{}/normal'.format(theme), section='appearance')[0]
        new_theme['cursorAccent'] = self.get_conf(
            '{}/ctrlclick'.format(theme), section='appearance')
        new_theme['selection'] = self.get_conf(
            '{}/occurrence'.format(theme), section='appearance')
        theme_colors = ANSI_COLORS[theme]
        for color in theme_colors:
            new_theme[color] = theme_colors[color]

        self.eval_javascript('setOption("{}", {})'.format('theme', new_theme))
        self.set_conf(
            'fontFamily', self.get_conf('font/family', section='appearance'))
        return new_theme

    def get_fonts(self):
        """List terminal CSS fonts."""
        return self.eval_javascript('getFonts()')

    def search_next(self, text, case=False, regex=False, word=False):
        """Search in the terminal for the given regex."""
        search_options = {'regex': regex,
                          'wholeWord': word,
                          'caseSensitive': case}
        return self.eval_javascript(
            'searchNext("{}",{})'.format(text, json.dumps(search_options)))

    def search_previous(self, text, case=False, regex=False, word=False):
        """Search in the terminal for the given regex."""
        search_options = {'regex': regex,
                          'wholeWord': word,
                          'caseSensitive': case}
        return self.eval_javascript(
            'searchPrevious("{}", {})'.format(text,
                                              json.dumps(search_options)))

    def exec_cmd(self, cmd):
        """Execute a command inside the terminal."""
        self.eval_javascript('exec("{0}")'.format(cmd))

    def exec_delayed_cmd(self, cmd, text_trigger, delay=1500):
        """
        Execute a command inside the terminal with a delay

        Parameters
        ----------
        cmd : str
            Command to type.
        text_trigger : str
            Text that should be in the console to trigger the command.
        delay : int, optional
            Milliseconds to wait to execute. The default is 1500.
        """
        self.eval_javascript(
            'exec_delayed("{0}", "{1}", {2})'.format(cmd, text_trigger, delay)
        )

    def reload(self):
        """Reload webview."""
        self.view.reload()

    def __alive_loopback(self):
        alive = self.is_alive()
        if not alive:
            self.terminal_closed.emit()
        else:
            QTimer.singleShot(250, self.__alive_loopback)

    def is_alive(self):
        """Check if terminal process is alive."""
        alive = self.eval_javascript('isAlive()')
        return alive

    def apply_zoom(self):
        zoom = self.get_conf('zoom')
        if zoom > 0:
            for __ in range(0, zoom):
                self.view.increase_font(new_term=True)
        if zoom < 0:
            for __ in range(0, -zoom):
                self.view.decrease_font(new_term=True)

    def set_option(self, option_name, option):
        """Set a configuration option in the terminal."""
        if type(option) == int:
            self.eval_javascript('setOption("{}", {})'.format(option_name,
                                                              option))
        else:
            self.eval_javascript('setOption("{}", "{}")'.format(option_name,
                                                                option))

    def apply_settings(self, options):
        """Apply custom settings given an option dictionary."""
        # Bell style option
        if 'sound' in options:
            bell_style = 'sound' if options['sound'] else 'none'
            self.set_option('bellStyle', bell_style)
        # Cursor option
        if 'cursor_type' in options:
            cursor_id = options['cursor_type']
            cursor_choices = {0: "block", 1: "underline", 2: "bar"}
            self.set_option('cursorStyle', cursor_choices[cursor_id])
        if 'buffer_limit' in options:
            new_lim = options['buffer_limit']
            self.set_option('scrollback', new_lim)
        if 'cursor_blink' in options:
            self.set_option('cursorBlink', int(options['cursor_blink']))

    def _apply_stylesheet(self, focus=False):
        """Apply stylesheet according to the current focus."""
        if focus:
            border_color = SpyderPalette.COLOR_ACCENT_3
        else:
            border_color = SpyderPalette.COLOR_BACKGROUND_4

        css = qstylizer.style.StyleSheet()
        css.QFrame.setValues(
            border=f'1px solid {border_color}',
            margin='0px 1px 0px 1px',
            padding='0px 0px 1px 0px',
            borderRadius='3px'
        )

        self.setStyleSheet(css.toString())


class TermView(QWebEngineView, SpyderWidgetMixin):
    """XTerm Wrapper."""
    sig_focus_in_event = Signal()
    """
    This signal is emitted when the widget receives focus.
    """

    sig_focus_out_event = Signal()
    """
    This signal is emitted when the widget loses focus.
    """

    def __init__(self, parent, term_url='http://127.0.0.1:8070', handler=None):
        """Webview main constructor."""
        super().__init__(parent, class_parent=parent)
        # Define a profile instance so each webview is independent even when
        # pointing to the same URL
        # Don't set parent to prevent warning when releasing profile
        self.profile = QWebEngineProfile()
        self.setPage(QWebEnginePage(self.profile, self))
        self.source_text = ''
        self.parent = parent
        self.channel = QWebChannel(self.page())
        self.page().setWebChannel(self.channel)
        self.channel.registerObject('handler', handler)

        self.term_url = QUrl(term_url)
        self.load(self.term_url)
        self.focusProxy().installEventFilter(self)

        self.document = self.page()
        try:
            self.document.profile().clearHttpCache()
        except AttributeError:
            pass

        self.initial_y_pos = 0
        self.setFocusPolicy(Qt.ClickFocus)
        self.setup()

    def setup(self):
        """Create the terminal context menu."""
        # Create context menu
        self.context_menu = self.get_menu(TermViewMenus.Context)
        for item in [self.get_action(TerminalMainWidgetActions.Copy),
                     self.get_action(TerminalMainWidgetActions.Paste),
                     self.get_action(TerminalMainWidgetActions.Clear)]:
            self.add_item_to_menu(
                item,
                menu=self.context_menu,
                section=TermViewSections.CommonActions,
            )

        for item in [self.get_action(TerminalMainWidgetActions.ZoomIn),
                     self.get_action(TerminalMainWidgetActions.ZoomOut)]:
            self.add_item_to_menu(
                item,
                menu=self.context_menu,
                section=TermViewSections.ZoomActions,
            )

    def copy(self):
        """Copy unicode text from terminal."""
        self.triggerPageAction(QWebEnginePage.Copy)

    def paste(self):
        """Paste unicode text into terminal."""
        clipboard = QApplication.clipboard()
        text = str(clipboard.text())
        if len(text.splitlines()) > 1:
            eol_chars = os.linesep
            text = eol_chars.join((text + eol_chars).splitlines())
        self.eval_javascript('pasteText({})'.format(repr(text)))

    def clear(self):
        """Clear the terminal."""
        self.eval_javascript('clearTerm()')

    def increase_font(self, new_term=False):
        """Increase terminal font."""
        if not new_term:
            zoom = self.get_conf('zoom')
            self.set_conf('zoom', zoom + 1)
        return self.eval_javascript('increaseFontSize()')

    def decrease_font(self, new_term=False):
        """Decrease terminal font."""
        if not new_term:
            zoom = self.get_conf('zoom')
            self.set_conf('zoom', zoom - 1)
        return self.eval_javascript('decreaseFontSize()')

    def contextMenuEvent(self, event):
        """Override Qt method."""
        self.context_menu.popup(event.globalPos())
        event.accept()

    def eval_javascript(self, script):
        """
        Evaluate Javascript instructions inside DOM with the expected prefix.
        """
        script = PREFIX + script
        self.document.runJavaScript("{}".format(script), self.return_js_value)

    def return_js_value(self, value):
        """Return the value of the function evaluated in Javascript."""
        return value

    def wheelEvent(self, event):
        """Catch and process wheel scrolling events via Javascript."""
        delta = event.angleDelta().y()
        self.eval_javascript('scrollTerm({0})'.format(delta))

    def event(self, event):
        """Grab all keyboard input."""
        if event.type() == QEvent.ShortcutOverride:
            self.keyPressEvent(event)
            return True
        return True

    def keyPressEvent(self, event):
        """Qt override method."""
        key = event.key()
        modifiers = event.modifiers()

        if modifiers & Qt.ShiftModifier:
            key += Qt.SHIFT
        if modifiers & Qt.ControlModifier:
            key += Qt.CTRL
        if modifiers & Qt.AltModifier:
            key += Qt.ALT
        if modifiers & Qt.MetaModifier:
            key += Qt.META

        sequence = QKeySequence(key).toString(QKeySequence.PortableText)
        if event == QKeySequence.Paste:
            self.paste()
        elif sequence == self.get_shortcut('copy'):
            self.copy()
        elif sequence == self.get_shortcut('paste'):
            self.paste()
        elif sequence == self.get_shortcut('clear'):
            self.clear()
        elif sequence == self.get_shortcut('zoom_in'):
            self.increase_font()
        elif sequence == self.get_shortcut('zoom_out'):
            self.decrease_font()
        else:
            super().keyPressEvent(event)

    def eventFilter(self, widget, event):
        """
        Handle events that affect the view.
        All events (e.g. focus in/out) reach the focus proxy, not this
        widget itself. That's why this event filter is necessary.
        """
        if self.focusProxy() is widget:
            if event.type() == QEvent.FocusIn:
                self.sig_focus_in_event.emit()
            elif event.type() == QEvent.FocusOut:
                self.sig_focus_out_event.emit()
        return super().eventFilter(widget, event)


def test():
    """Plugin visual test."""
    from spyder.utils.qthelpers import qapplication
    app = qapplication(test_time=8)
    term = TerminalWidget(None)
    # term.resize(900, 700)
    term.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Setup script for spyder_terminal."""

# Standard library imports
import ast
import os
import os.path as osp
import sys

# Third party imports
from setuptools import find_packages, setup

from setupbase import (BuildStatic, CleanComponents, COMPONENTS, HERE,
                       SdistWithBuildStatic)


def get_version(module='spyder_terminal'):
    """Get version."""
    with open(os.path.join(HERE, module, '__init__.py'), 'r') as f:
        data = f.read()
    lines = data.split('\n')
    for line in lines:
        if line.startswith('VERSION_INFO'):
            version_tuple = ast.literal_eval(line.split('=')[-1].strip())
            version = '.'.join(map(str, version_tuple))
            break
    return version


def get_description():
    """Get long description."""
    with open(os.path.join(HERE, 'README.rst'), 'r') as f:
        data = f.read()
    return data


REQUIREMENTS = [
    'spyder>=6.1.0,<7.0.0',
    'tornado',
    'terminado>=0.13.1',
    'coloredlogs',
    'requests'
]

EXTRAS_REQUIRE = {
    'test': [
        'pytest',
        'pytest-cov',
        'flaky',
        'pytest-qt',
        'pytest-timeout'
    ]
}

# Verify that COMPONENTS exist before trying to build wheels
if any([arg == 'bdist_wheel' for arg in sys.argv]):
    if not osp.isdir(COMPONENTS):
        print("\nWARNING: Server components are missing!! Please run "
              "'python setup.py sdist' first.\n")
        sys.exit(1)

cmdclass = {
    'build_static': BuildStatic,
    'sdist': SdistWithBuildStatic,
    'clean_components': CleanComponents
}


setup(
    name='spyder-terminal',
    version=get_version(),
    cmdclass=cmdclass,
    keywords=['Spyder', 'Plugin'],
    url='https://github.com/spyder-ide/spyder-terminal',
    license='MIT',
    author='Spyder Project Contributors',
    author_email='spyder.python@gmail.com',
    description='Spyder Plugin for displaying a virtual terminal '
                '(OS independent) inside the main Spyder window',
    long_description=get_description(),
    long_description_content_type='text/x-rst',
    packages=find_packages(exclude=['contrib', 'docs']),
    install_requires=REQUIREMENTS,
    python_requires='>=3.9',
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        ],
    extras_require=EXTRAS_REQUIRE,
    entry_points={
        "spyder.plugins": [
            "terminal = spyder_terminal.terminalplugin:TerminalPlugin"
        ],
    })
